# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""A TF-GAN-backed GAN Estimator that works on TPU."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.contrib.gan.python import namedtuples as tfgan_tuples
from tensorflow.contrib.gan.python import train as tfgan_train
from tensorflow.contrib.gan.python.estimator.python import gan_estimator_impl as gan_estimator_lib
from tensorflow.contrib.tpu.python.tpu import tpu_estimator
from tensorflow.contrib.tpu.python.tpu import tpu_optimizer
from tensorflow.contrib.training.python.training import training
from tensorflow.python.estimator import model_fn as model_fn_lib
from tensorflow.python.framework import ops
from tensorflow.python.ops import control_flow_ops
from tensorflow.python.ops import metrics as metrics_lib
from tensorflow.python.ops.losses import losses

__all__ = [
    'TPUGANEstimator',
]


class TPUGANEstimator(tpu_estimator.TPUEstimator):
  """An estimator for Generative Adversarial Networks (GANs) on TPU.

  This Estimator is backed by TFGAN. It is similar to `tfgan.GANEstimator`,
  but works on TPU.

  Example:

  ```python
      import tensorflow as tf
      tfgan = tf.contrib.gan

      # See TFGAN's `train.py` for a description of the generator and
      # discriminator API.
      def generator_fn(generator_inputs):
        ...
        return generated_data

      def discriminator_fn(data, conditioning):
        ...
        return logits

      # Create GAN estimator.
      config = tpu_config.RunConfig(model_dir='/my/dir')
      gan_estimator = tfgan.estimator.TPUGANEstimator(
          generator_fn=generator_fn,
          discriminator_fn=discriminator_fn,
          generator_loss_fn=tfgan.losses.wasserstein_generator_loss,
          discriminator_loss_fn=tfgan.losses.wasserstein_discriminator_loss,
          generator_optimizer=tf.compat.v1.train.AdamOptimizer(0.1, 0.5),
          discriminator_optimizer=tf.compat.v1.train.AdamOptimizer(0.1, 0.5),
          train_batch_size=4,
          config=config)

      # Train estimator.
      gan_estimator.train(train_input_fn, train_steps)

      # Evaluate resulting estimator.
      gan_estimator.evaluate(eval_input_fn, eval_steps)

      # Generate samples from generator.
      predictions = np.array([
          x['generated_data'] for x in gan_estimator.predict(predict_input_fn)])
  ```
  """

  def __init__(self,
               # Arguments to construct the `model_fn`.
               generator_fn=None,
               discriminator_fn=None,
               generator_loss_fn=None,
               discriminator_loss_fn=None,
               generator_optimizer=None,
               discriminator_optimizer=None,
               get_eval_metric_ops_fn=None,
               add_summaries=None,
               joint_train=False,
               gan_train_steps=tfgan_tuples.GANTrainSteps(1, 1),
               # TPUEstimator options.
               model_dir=None,
               config=None,
               params=None,
               use_tpu=True,
               train_batch_size=None,
               eval_batch_size=None,
               predict_batch_size=None,
               batch_axis=None,
               eval_on_tpu=True,
               export_to_tpu=True,
               warm_start_from=None):
    """Initializes a TPUGANEstimator instance.

    Args:
      generator_fn: A python function that takes a Tensor, Tensor list, or
        Tensor dictionary as inputs and returns the outputs of the GAN
        generator. See `TFGAN` for more details and examples. Additionally, if
        it has an argument called `mode`, the Estimator's `mode` will be passed
        in (ex TRAIN, EVAL, PREDICT). This is useful for things like batch
        normalization.
      discriminator_fn: A python function that takes the output of
        `generator_fn` or real data in the GAN setup, and `generator_inputs`.
        Outputs a Tensor in the range [-inf, inf]. See `TFGAN` for more details
        and examples.
      generator_loss_fn: The loss function on the generator. Takes a `GANModel`
        tuple.
      discriminator_loss_fn: The loss function on the discriminator. Takes a
        `GANModel` tuple.
      generator_optimizer: The optimizer for generator updates, or a function
        that takes no arguments and returns an optimizer. This function will
        be called when the default graph is the `GANEstimator`'s graph, so
        utilities like `tf.contrib.framework.get_or_create_global_step` will
        work.
      discriminator_optimizer: Same as `generator_optimizer`, but for the
        discriminator updates.
      get_eval_metric_ops_fn: A function that takes a list of arguments and
        returns a dict of metric results keyed by name. The output of this
        function is passed into `tf.estimator.EstimatorSpec` during evaluation.
        The arguments must be:
            * generator_inputs
            * generated_data
            * real_data
            * discriminator_real_outputs
            * discriminator_gen_outputs
      add_summaries: `None`, a single `SummaryType`, or a list of `SummaryType`.
        This is ignored for jobs that run on TPU, such as the train job if
        `use_tpu` is `True` or the eval job if `eval_on_tpu` is `True`.
      joint_train: A Python boolean. If `True`, jointly train the generator and
        the discriminator. If `False`, sequentially train them. See `train.py`
        in TFGAN for more details on the differences between the two GAN
        training methods.
      gan_train_steps: A `tfgan.GANTrainSteps` named tuple describing the ratio
        of generator to discriminator steps. For now, only supports 1:1
        training.
      model_dir: Same as `TPUEstimator`: Directory to save model parameters,
        graph and etc. This can also be used to load checkpoints from the
        directory into a estimator to continue training a previously saved
        model. If `None`, the model_dir in `config` will be used if set. If both
        are set, they must be same. If both are `None`, a temporary directory
        will be used.
      config: Same as `TPUEstimator`: An `tpu_config.RunConfig` configuration
        object. Cannot be `None`.
      params: Same as `TPUEstimator`: An optional `dict` of hyper parameters
        that will be passed into `input_fn` and `model_fn`.  Keys are names of
        parameters, values are basic python types. There are reserved keys for
        `TPUEstimator`, including 'batch_size'.
      use_tpu: Same as `TPUEstimator`: A bool indicating whether TPU support is
        enabled. Currently, TPU training and evaluation respect this bit, but
        eval_on_tpu can override execution of eval. See below. Predict still
        happens on CPU.
      train_batch_size: Same as `TPUEstimator`: An int representing the global
        training batch size. TPUEstimator transforms this global batch size to a
        per-shard batch size, as params['batch_size'], when calling `input_fn`
        and `model_fn`. Cannot be `None` if `use_tpu` is `True`. Must be
        divisible by total number of replicas.
      eval_batch_size: Same as `TPUEstimator`: An int representing evaluation
        batch size. Must be divisible by total number of replicas.
      predict_batch_size: Same as `TPUEstimator`: An int representing the
        prediction batch size. Must be divisible by total number of replicas.
      batch_axis: Same as `TPUEstimator`: A python tuple of int values
        describing how each tensor produced by the Estimator `input_fn` should
        be split across the TPU compute shards. For example, if your input_fn
        produced (images, labels) where the images tensor is in `HWCN` format,
        your shard dimensions would be [3, 0], where 3 corresponds to the `N`
        dimension of your images Tensor, and 0 corresponds to the dimension
        along which to split the labels to match up with the corresponding
        images. If None is supplied, and per_host_input_for_training is True,
        batches will be sharded based on the major dimension. If
        tpu_config.per_host_input_for_training is False or `PER_HOST_V2`,
        batch_axis is ignored.
      eval_on_tpu: Same as `TPUEstimator`: If False, evaluation runs on CPU or
        GPU. In this case, the model_fn must return `EstimatorSpec` when called
        with `mode` as `EVAL`.
      export_to_tpu: Same as `TPUEstimator`: If True, `export_savedmodel()`
        exports a metagraph for serving on TPU besides the one on CPU.
      warm_start_from: Same as `TPUEstimator`: Optional string filepath to a
        checkpoint or SavedModel to warm-start from, or a
        `tf.estimator.WarmStartSettings` object to fully configure
        warm-starting.  If the string filepath is provided instead of a
        `WarmStartSettings`, then all variables are warm-started, and it is
        assumed that vocabularies and Tensor names are unchanged.

    Raises:
      ValueError: If loss functions aren't callable.
      ValueError: If `gan_train_steps` isn't a `tfgan_tuples.GANTrainSteps`
        tuple.
      ValueError: If `gan_train_steps` isn't 1:1 training.
    """
    if not callable(generator_loss_fn):
      raise ValueError('generator_loss_fn must be callable.')
    if not callable(discriminator_loss_fn):
      raise ValueError('discriminator_loss_fn must be callable.')
    if not isinstance(gan_train_steps, tfgan_tuples.GANTrainSteps):
      raise ValueError(
          '`gan_train_steps` must be `tfgan_tuples.GANTrainSteps`. Instead, '
          'was type: %s' % type(gan_train_steps))
    if (gan_train_steps.generator_train_steps != 1 or
        gan_train_steps.discriminator_train_steps != 1):
      raise ValueError('Estimator currently only supports 1:1 training.')

    if use_tpu:
      generator_optimizer = _maybe_make_cross_shard_optimizer(
          generator_optimizer)
      discriminator_optimizer = _maybe_make_cross_shard_optimizer(
          discriminator_optimizer)

    def _model_fn(features, labels, mode, params):
      """GANEstimator model function."""
      del params  # unused
      if mode not in [model_fn_lib.ModeKeys.TRAIN, model_fn_lib.ModeKeys.EVAL,
                      model_fn_lib.ModeKeys.PREDICT]:
        raise ValueError('Mode not recognized: %s' % mode)
      real_data = labels  # rename inputs for clarity
      generator_inputs = features  # rename inputs for clarity

      # Make GANModel, which encapsulates the GAN model architectures.
      # TODO(joelshor): Switch TF-GAN over to TPU-compatible summaries, then
      # remove `add_summaries` logic below.
      is_on_tpu = _is_on_tpu(mode, use_tpu, eval_on_tpu)
      gan_model = gan_estimator_lib._get_gan_model(  # pylint:disable=protected-access
          mode, generator_fn, discriminator_fn, real_data, generator_inputs,
          add_summaries=None if is_on_tpu else add_summaries)

      # Make the TPUEstimatorSpec, which incorporates the GANModel, losses, eval
      # metrics, and optimizers (if required).
      estimator_spec = _get_estimator_spec(
          mode, gan_model, generator_loss_fn, discriminator_loss_fn,
          get_eval_metric_ops_fn, generator_optimizer, discriminator_optimizer,
          joint_train, is_on_tpu, gan_train_steps)
      assert isinstance(estimator_spec, tpu_estimator.TPUEstimatorSpec)
      return estimator_spec

    super(TPUGANEstimator, self).__init__(
        model_fn=_model_fn,
        model_dir=model_dir,
        config=config,
        params=params,
        use_tpu=use_tpu,
        train_batch_size=train_batch_size,
        eval_batch_size=eval_batch_size,
        predict_batch_size=predict_batch_size,
        batch_axis=batch_axis,
        eval_on_tpu=eval_on_tpu,
        export_to_tpu=export_to_tpu,
        warm_start_from=warm_start_from)


def _is_on_tpu(mode, use_tpu, eval_on_tpu):
  if mode == model_fn_lib.ModeKeys.TRAIN:
    return use_tpu
  elif mode == model_fn_lib.ModeKeys.EVAL:
    return eval_on_tpu
  else:
    return False


def _get_estimator_spec(
    mode, gan_model, generator_loss_fn, discriminator_loss_fn,
    get_eval_metric_ops_fn, generator_optimizer, discriminator_optimizer,
    joint_train, is_on_tpu, gan_train_steps):
  """Get the TPUEstimatorSpec for the current mode."""
  if mode == model_fn_lib.ModeKeys.PREDICT:
    estimator_spec = tpu_estimator.TPUEstimatorSpec(
        mode=mode, predictions={'generated_data': gan_model.generated_data})
  elif mode == model_fn_lib.ModeKeys.EVAL:
    gan_loss = tfgan_tuples.GANLoss(
        generator_loss=generator_loss_fn(
            gan_model, add_summaries=not is_on_tpu),
        discriminator_loss=discriminator_loss_fn(
            gan_model, add_summaries=not is_on_tpu))
    # Eval losses for metrics must preserve batch dimension.
    gan_loss_no_reduction = tfgan_tuples.GANLoss(
        generator_loss=generator_loss_fn(
            gan_model, add_summaries=False, reduction=losses.Reduction.NONE),
        discriminator_loss=discriminator_loss_fn(
            gan_model, add_summaries=False, reduction=losses.Reduction.NONE))
    estimator_spec = _get_eval_estimator_spec(
        gan_model, gan_loss, gan_loss_no_reduction, get_eval_metric_ops_fn)
  else:  # model_fn_lib.ModeKeys.TRAIN:
    gan_loss = tfgan_tuples.GANLoss(
        generator_loss=generator_loss_fn(
            gan_model, add_summaries=not is_on_tpu),
        discriminator_loss=discriminator_loss_fn(
            gan_model, add_summaries=not is_on_tpu))

    # Construct optimizers if arguments were callable. For TPUs, they must be
    # `CrossShardOptimizer`.
    g_callable = callable(generator_optimizer)
    gopt = generator_optimizer() if g_callable  else generator_optimizer
    d_callable = callable(discriminator_optimizer)
    dopt = discriminator_optimizer() if d_callable else discriminator_optimizer

    estimator_spec = _get_train_estimator_spec(
        gan_model, gan_loss, gopt, dopt, joint_train, gan_train_steps)

  return estimator_spec


def _get_eval_estimator_spec(gan_model, gan_loss, gan_loss_no_reduction,
                             get_eval_metric_ops_fn):
  """Return an TPUEstimatorSpec for the eval case."""
  # Make the metric function and tensor names.
  if get_eval_metric_ops_fn is not None:
    def metric_fn(
        generator_inputs, generated_data, real_data, discriminator_real_outputs,
        discriminator_gen_outputs, generator_loss, discriminator_loss):
      """`metric_fn` used in TPUEstimator to calculate metrics."""
      eval_metric_ops = {
          'generator_loss': metrics_lib.mean(generator_loss),
          'discriminator_loss': metrics_lib.mean(discriminator_loss),
      }
      custom_eval_metric_ops = get_eval_metric_ops_fn(
          generator_inputs, generated_data, real_data,
          discriminator_real_outputs, discriminator_gen_outputs)
      if not isinstance(custom_eval_metric_ops, dict):
        raise TypeError('`get_eval_metric_ops_fn` must return a dict, '
                        'received: {}'.format(custom_eval_metric_ops))
      eval_metric_ops.update(custom_eval_metric_ops)
      return eval_metric_ops
    tensors = {
        'generator_loss': gan_loss_no_reduction.generator_loss,
        'discriminator_loss': gan_loss_no_reduction.discriminator_loss,
        'generator_inputs': gan_model.generator_inputs,
        'generated_data': gan_model.generated_data,
        'real_data': gan_model.real_data,
        'discriminator_real_outputs': gan_model.discriminator_real_outputs,
        'discriminator_gen_outputs': gan_model.discriminator_gen_outputs,
    }
  else:
    def metric_fn(generator_loss, discriminator_loss):
      return {
          'generator_loss': metrics_lib.mean(generator_loss),
          'discriminator_loss': metrics_lib.mean(discriminator_loss),
      }
    tensors = {
        'generator_loss': gan_loss_no_reduction.generator_loss,
        'discriminator_loss': gan_loss_no_reduction.discriminator_loss,
    }

  scalar_loss = gan_loss.generator_loss + gan_loss.discriminator_loss
  return tpu_estimator.TPUEstimatorSpec(
      mode=model_fn_lib.ModeKeys.EVAL,
      predictions=gan_model.generated_data,
      loss=scalar_loss,
      eval_metrics=(metric_fn, tensors))


def _get_train_estimator_spec(
    gan_model, gan_loss, generator_optimizer, discriminator_optimizer,
    joint_train, gan_train_steps):
  """Return a TPUEstimatorSpec for the train case."""
  scalar_loss = gan_loss.generator_loss + gan_loss.discriminator_loss

  # Get generator and discriminator update ops. We split them so that update
  # ops aren't accidentally run multiple times. For now, throw an error if
  # there are update ops that aren't associated with either the generator or
  # the discriminator. Might modify the `kwargs` dictionary.
  gen_update_ops, dis_update_ops = tfgan_train._get_update_ops(  # pylint:disable=protected-access
      {}, gan_model.generator_scope.name, gan_model.discriminator_scope.name)

  def gen_train_op():
    with ops.name_scope('generator_train'):
      return training.create_train_op(
          total_loss=gan_loss.generator_loss,
          optimizer=generator_optimizer,
          variables_to_train=gan_model.generator_variables,
          update_ops=gen_update_ops)
  def dis_train_op():
    with ops.name_scope('discriminator_train'):
      return training.create_train_op(
          total_loss=gan_loss.discriminator_loss,
          optimizer=discriminator_optimizer,
          variables_to_train=gan_model.discriminator_variables,
          update_ops=dis_update_ops)

  # Either optimize the generator and discriminator sequentially or jointly.
  tpu_train_op = _combine_train_ops(gen_train_op, dis_train_op, joint_train,
                                    gan_train_steps)

  return tpu_estimator.TPUEstimatorSpec(
      loss=scalar_loss,
      mode=model_fn_lib.ModeKeys.TRAIN,
      train_op=tpu_train_op)


# TODO(joelshor): Add support for multiple D / G steps.
def _combine_train_ops(gen_train_op, dis_train_op, joint_train,
                       gan_train_steps):
  """Combine generator and discriminator train ops into a single op."""
  del gan_train_steps
  if joint_train:
    tpu_train_op = control_flow_ops.group(gen_train_op(), dis_train_op(),
                                          name='joint_train')
  else:
    with ops.control_dependencies([dis_train_op()]):
      tpu_train_op = gen_train_op()

  return tpu_train_op


def _maybe_make_cross_shard_optimizer(opt):
  if callable(opt):
    if not isinstance(opt(), tpu_optimizer.CrossShardOptimizer):
      return lambda: tpu_optimizer.CrossShardOptimizer(opt())
  elif not isinstance(opt, tpu_optimizer.CrossShardOptimizer):
    return tpu_optimizer.CrossShardOptimizer(opt)
  return opt



# This file is MACHINE GENERATED! Do not edit.
# Generated by: tensorflow/python/tools/api/generator/create_python_api.py script.
"""Math Operations.

Note: Functions taking `Tensor` arguments can also take anything accepted by
`tf.convert_to_tensor`.

Note: Elementwise binary operations in TensorFlow follow [numpy-style
broadcasting](http://docs.scipy.org/doc/numpy/user/basics.broadcasting.html).

TensorFlow provides a variety of math functions including:

* Basic arithmetic operators and trigonometric functions.
* Special math functions (like: `tf.math.igamma` and `tf.math.zeta`)
* Complex number functions (like: `tf.math.imag` and `tf.math.angle`)
* Reductions and scans (like: `tf.math.reduce_mean` and `tf.math.cumsum`)
* Segment functions (like: `tf.math.segment_sum`)

See: `tf.linalg` for matrix and tensor functions.

<a id=Segmentation></a>

## About Segmentation

TensorFlow provides several operations that you can use to perform common
math computations on tensor segments.
Here a segmentation is a partitioning of a tensor along
the first dimension, i.e. it  defines a mapping from the first dimension onto
`segment_ids`. The `segment_ids` tensor should be the size of
the first dimension, `d0`, with consecutive IDs in the range `0` to `k`,
where `k<d0`.
In particular, a segmentation of a matrix tensor is a mapping of rows to
segments.

For example:

```python
c = tf.constant([[1,2,3,4], [-1,-2,-3,-4], [5,6,7,8]])
tf.math.segment_sum(c, tf.constant([0, 0, 1]))
#  ==>  [[0 0 0 0]
#        [5 6 7 8]]
```

The standard `segment_*` functions assert that the segment indices are sorted.
If you have unsorted indices use the equivalent `unsorted_segment_` function.
Thses functions take an additional argument `num_segments` so that the output
tensor can be efficiently allocated.

``` python
c = tf.constant([[1,2,3,4], [-1,-2,-3,-4], [5,6,7,8]])
tf.math.unsorted_segment_sum(c, tf.constant([0, 1, 0]), num_segments=2)
# ==> [[ 6,  8, 10, 12],
#       [-1, -2, -3, -4]]
```


"""

from __future__ import print_function as _print_function

from tensorflow.python import abs
from tensorflow.python import accumulate_n
from tensorflow.python import acos
from tensorflow.python import acosh
from tensorflow.python import add
from tensorflow.python import add_n
from tensorflow.python import angle
from tensorflow.python import argmax
from tensorflow.python import argmin
from tensorflow.python import asin
from tensorflow.python import asinh
from tensorflow.python import atan
from tensorflow.python import atan2
from tensorflow.python import atanh
from tensorflow.python import bessel_i0
from tensorflow.python import bessel_i0e
from tensorflow.python import bessel_i1
from tensorflow.python import bessel_i1e
from tensorflow.python import betainc
from tensorflow.python import bincount_v1 as bincount
from tensorflow.python import ceil
from tensorflow.python import conj
from tensorflow.python import cos
from tensorflow.python import cosh
from tensorflow.python import count_nonzero
from tensorflow.python import cumprod
from tensorflow.python import cumsum
from tensorflow.python import digamma
from tensorflow.python import div_no_nan as divide_no_nan
from tensorflow.python import divide
from tensorflow.python import equal
from tensorflow.python import erf
from tensorflow.python import erfc
from tensorflow.python import exp
from tensorflow.python import expm1
from tensorflow.python import floor
from tensorflow.python import floor_mod as floormod
from tensorflow.python import floor_mod as mod
from tensorflow.python import floordiv
from tensorflow.python import greater
from tensorflow.python import greater_equal
from tensorflow.python import igamma
from tensorflow.python import igammac
from tensorflow.python import imag
from tensorflow.python import invert_permutation
from tensorflow.python import is_finite
from tensorflow.python import is_inf
from tensorflow.python import is_nan
from tensorflow.python import is_non_decreasing
from tensorflow.python import is_strictly_increasing
from tensorflow.python import lbeta
from tensorflow.python import less
from tensorflow.python import less_equal
from tensorflow.python import lgamma
from tensorflow.python import log
from tensorflow.python import log1p
from tensorflow.python import log_sigmoid
from tensorflow.python import logical_and
from tensorflow.python import logical_not
from tensorflow.python import logical_or
from tensorflow.python import logical_xor
from tensorflow.python import maximum
from tensorflow.python import minimum
from tensorflow.python import multiply
from tensorflow.python import multiply_no_nan
from tensorflow.python import neg as negative
from tensorflow.python import next_after as nextafter
from tensorflow.python import not_equal
from tensorflow.python import polygamma
from tensorflow.python import polyval
from tensorflow.python import pow
from tensorflow.python import real
from tensorflow.python import reciprocal
from tensorflow.python import reduce_all_v1 as reduce_all
from tensorflow.python import reduce_any_v1 as reduce_any
from tensorflow.python import reduce_euclidean_norm
from tensorflow.python import reduce_logsumexp_v1 as reduce_logsumexp
from tensorflow.python import reduce_max_v1 as reduce_max
from tensorflow.python import reduce_mean_v1 as reduce_mean
from tensorflow.python import reduce_min_v1 as reduce_min
from tensorflow.python import reduce_prod_v1 as reduce_prod
from tensorflow.python import reduce_std
from tensorflow.python import reduce_sum_v1 as reduce_sum
from tensorflow.python import reduce_variance
from tensorflow.python import rint
from tensorflow.python import round
from tensorflow.python import rsqrt
from tensorflow.python import scalar_mul
from tensorflow.python import segment_max
from tensorflow.python import segment_mean
from tensorflow.python import segment_min
from tensorflow.python import segment_prod
from tensorflow.python import segment_sum
from tensorflow.python import sigmoid
from tensorflow.python import sign
from tensorflow.python import sin
from tensorflow.python import sinh
from tensorflow.python import sqrt
from tensorflow.python import square
from tensorflow.python import squared_difference
from tensorflow.python import subtract
from tensorflow.python import tan
from tensorflow.python import tanh
from tensorflow.python import truediv
from tensorflow.python import unsorted_segment_max
from tensorflow.python import unsorted_segment_mean
from tensorflow.python import unsorted_segment_min
from tensorflow.python import unsorted_segment_prod
from tensorflow.python import unsorted_segment_sqrt_n
from tensorflow.python import unsorted_segment_sum
from tensorflow.python import xdivy
from tensorflow.python import xlogy
from tensorflow.python import zeta
from tensorflow.python.ops.confusion_matrix import confusion_matrix_v1 as confusion_matrix
from tensorflow.python.ops.gen_nn_ops import softplus
from tensorflow.python.ops.gen_nn_ops import softsign
from tensorflow.python.ops.nn import in_top_k
from tensorflow.python.ops.nn import l2_normalize
from tensorflow.python.ops.nn import log_softmax
from tensorflow.python.ops.nn import softmax
from tensorflow.python.ops.nn import top_k
from tensorflow.python.ops.nn import zero_fraction

del _print_function

import sys as _sys
from tensorflow.python.util import deprecation_wrapper as _deprecation_wrapper

if not isinstance(_sys.modules[__name__], _deprecation_wrapper.DeprecationWrapper):
  _sys.modules[__name__] = _deprecation_wrapper.DeprecationWrapper(
      _sys.modules[__name__], "math")



# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Ftrl-proximal for TensorFlow."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorflow.python.keras.optimizer_v2 import optimizer_v2
from tensorflow.python.ops import init_ops
from tensorflow.python.ops import math_ops
from tensorflow.python.training import training_ops
from tensorflow.python.util.tf_export import keras_export


@keras_export('keras.optimizers.Ftrl')
class Ftrl(optimizer_v2.OptimizerV2):
  r"""Optimizer that implements the FTRL algorithm.

  See Algorithm 1 of this [paper](
  https://www.eecs.tufts.edu/~dsculley/papers/ad-click-prediction.pdf).
  This version has support for both online L2 (the L2 penalty given in the paper
  above) and shrinkage-type L2 (which is the addition of an L2 penalty to the
  loss function).

  Initialization:
  $$t = 0$$
  $$n_{0} = 0$$
  $$\sigma_{0} = 0$$
  $$z_{0} = 0$$

  Update ($$i$$ is variable index):
  $$t = t + 1$$
  $$n_{t,i} = n_{t-1,i} + g_{t,i}^{2}$$
  $$\sigma_{t,i} = (\sqrt{n_{t,i}} - \sqrt{n_{t-1,i}}) / \alpha$$
  $$z_{t,i} = z_{t-1,i} + g_{t,i} - \sigma_{t,i} * w_{t,i}$$
  $$w_{t,i} = - ((\beta+\sqrt{n+{t}}) / \alpha + \lambda_{2})^{-1} * (z_{i} -
               sgn(z_{i}) * \lambda_{1}) if \abs{z_{i}} > \lambda_{i} else 0$$

  Check the documentation for the l2_shrinkage_regularization_strength
  parameter for more details when shrinkage is enabled, where gradient is
  replaced with gradient_with_shrinkage.
  """

  def __init__(self,
               learning_rate,
               learning_rate_power=-0.5,
               initial_accumulator_value=0.1,
               l1_regularization_strength=0.0,
               l2_regularization_strength=0.0,
               name='Ftrl',
               l2_shrinkage_regularization_strength=0.0,
               **kwargs):
    r"""Construct a new FTRL optimizer.

    Args:
      learning_rate: A float value or a constant float `Tensor`.
      learning_rate_power: A float value, must be less or equal to zero.
        Controls how the learning rate decreases during training. Use zero for
        a fixed learning rate.
      initial_accumulator_value: The starting value for accumulators.
        Only zero or positive values are allowed.
      l1_regularization_strength: A float value, must be greater than or
        equal to zero.
      l2_regularization_strength: A float value, must be greater than or
        equal to zero.
      name: Optional name prefix for the operations created when applying
        gradients.  Defaults to "Ftrl".
      l2_shrinkage_regularization_strength: A float value, must be greater than
        or equal to zero. This differs from L2 above in that the L2 above is a
        stabilization penalty, whereas this L2 shrinkage is a magnitude penalty.
        The FTRL formulation can be written as:
        w_{t+1} = argmin_w(\hat{g}_{1:t}w + L1*||w||_1 + L2*||w||_2^2), where
        \hat{g} = g + (2*L2_shrinkage*w), and g is the gradient of the loss
        function w.r.t. the weights w.
        Specifically, in the absence of L1 regularization, it is equivalent to
        the following update rule:
        w_{t+1} = w_t - lr_t / (1 + 2*L2*lr_t) * g_t -
                  2*L2_shrinkage*lr_t / (1 + 2*L2*lr_t) * w_t
        where lr_t is the learning rate at t.
        When input is sparse shrinkage will only happen on the active weights.\
      **kwargs: keyword arguments. Allowed to be {`clipnorm`, `clipvalue`, `lr`,
        `decay`}. `clipnorm` is clip gradients by norm; `clipvalue` is clip
        gradients by value, `decay` is included for backward compatibility to
        allow time inverse decay of learning rate. `lr` is included for backward
        compatibility, recommended to use `learning_rate` instead.

    Raises:
      ValueError: If one of the arguments is invalid.

    References
      See [paper]
        (https://www.eecs.tufts.edu/~dsculley/papers/ad-click-prediction.pdf)
    """
    super(Ftrl, self).__init__(name, **kwargs)

    if initial_accumulator_value < 0.0:
      raise ValueError(
          'initial_accumulator_value %f needs to be positive or zero' %
          initial_accumulator_value)
    if learning_rate_power > 0.0:
      raise ValueError('learning_rate_power %f needs to be negative or zero' %
                       learning_rate_power)
    if l1_regularization_strength < 0.0:
      raise ValueError(
          'l1_regularization_strength %f needs to be positive or zero' %
          l1_regularization_strength)
    if l2_regularization_strength < 0.0:
      raise ValueError(
          'l2_regularization_strength %f needs to be positive or zero' %
          l2_regularization_strength)
    if l2_shrinkage_regularization_strength < 0.0:
      raise ValueError(
          'l2_shrinkage_regularization_strength %f needs to be positive'
          ' or zero' % l2_shrinkage_regularization_strength)

    self._set_hyper('learning_rate', learning_rate)
    self._set_hyper('decay', self._initial_decay)
    self._set_hyper('learning_rate_power', learning_rate_power)
    self._set_hyper('l1_regularization_strength', l1_regularization_strength)
    self._set_hyper('l2_regularization_strength', l2_regularization_strength)
    self._initial_accumulator_value = initial_accumulator_value
    self._l2_shrinkage_regularization_strength = (
        l2_shrinkage_regularization_strength)

  def _create_slots(self, var_list):
    # Create the "accum" and "linear" slots.
    for var in var_list:
      dtype = var.dtype.base_dtype
      init = init_ops.constant_initializer(
          self._initial_accumulator_value, dtype=dtype)
      self.add_slot(var, 'accumulator', init)
      self.add_slot(var, 'linear')

  def _resource_apply_dense(self, grad, var):
    var_dtype = var.dtype.base_dtype
    lr_t = self._decayed_lr(var_dtype)
    learning_rate_power = self._get_hyper('learning_rate_power', var_dtype)
    l1_regularization_strength = self._get_hyper('l1_regularization_strength',
                                                 var_dtype)
    l2_regularization_strength = self._get_hyper('l2_regularization_strength',
                                                 var_dtype)
    accum = self.get_slot(var, 'accumulator')
    linear = self.get_slot(var, 'linear')
    if self._l2_shrinkage_regularization_strength <= 0.0:
      return training_ops.resource_apply_ftrl(
          var.handle,
          accum.handle,
          linear.handle,
          grad,
          lr_t,
          l1_regularization_strength,
          l2_regularization_strength,
          learning_rate_power,
          use_locking=self._use_locking)
    else:
      return training_ops.resource_apply_ftrl_v2(
          var.handle,
          accum.handle,
          linear.handle,
          grad,
          lr_t,
          l1_regularization_strength,
          l2_regularization_strength,
          math_ops.cast(self._l2_shrinkage_regularization_strength, var_dtype),
          learning_rate_power,
          use_locking=self._use_locking)

  def _resource_apply_sparse(self, grad, var, indices):
    var_dtype = var.dtype.base_dtype
    lr_t = self._decayed_lr(var_dtype)
    learning_rate_power = self._get_hyper('learning_rate_power', var_dtype)
    l1_regularization_strength = self._get_hyper('l1_regularization_strength',
                                                 var_dtype)
    l2_regularization_strength = self._get_hyper('l2_regularization_strength',
                                                 var_dtype)
    accum = self.get_slot(var, 'accumulator')
    linear = self.get_slot(var, 'linear')
    if self._l2_shrinkage_regularization_strength <= 0.0:
      return training_ops.resource_sparse_apply_ftrl(
          var.handle,
          accum.handle,
          linear.handle,
          grad,
          indices,
          lr_t,
          l1_regularization_strength,
          l2_regularization_strength,
          learning_rate_power,
          use_locking=self._use_locking)
    else:
      return training_ops.resource_sparse_apply_ftrl_v2(
          var.handle,
          accum.handle,
          linear.handle,
          grad,
          indices,
          lr_t,
          l1_regularization_strength,
          l2_regularization_strength,
          math_ops.cast(self._l2_shrinkage_regularization_strength, var_dtype),
          learning_rate_power,
          use_locking=self._use_locking)

  def get_config(self):
    config = super(Ftrl, self).get_config()
    config.update({
        'learning_rate':
            self._serialize_hyperparameter('learning_rate'),
        'decay':
            self._serialize_hyperparameter('decay'),
        'initial_accumulator_value':
            self._initial_accumulator_value,
        'learning_rate_power':
            self._serialize_hyperparameter('learning_rate_power'),
        'l1_regularization_strength':
            self._serializer_hyperparameter('l1_regularization_strength'),
        'l2_regularization_strength':
            self._serializer_hyperparameter('l2_regularization_strength'),
        'l2_shrinkage_regularization_strength':
            self._l2_shrinkage_regularization_strength,
    })
    return config
